@@ -77,7 +77,6 @@
 )]
 // END - Embark standard lints v0.4
 // crate-specific exceptions:
-#![deny(missing_crate_level_docs)]
 #![allow(clippy::float_cmp, clippy::manual_range_contains)]
 
 mod flamegraph;
@@ -87,7 +86,10 @@ pub use {egui, puffin};
 
 use egui::*;
 use puffin::*;
-use std::sync::{Arc, Mutex};
+use std::{
+    collections::{BTreeMap, BTreeSet},
+    sync::{Arc, Mutex},
+};
 
 const ERROR_COLOR: Color32 = Color32::RED;
 const HOVER_COLOR: Rgba = Rgba::from_rgb(0.8, 0.8, 0.8);
@@ -109,7 +111,7 @@ pub fn profiler_window(ctx: &egui::CtxRef) -> bool {
     open
 }
 
-static PROFILE_UI: once_cell::sync::Lazy<Mutex<ProfilerUi>> =
+static PROFILE_UI: once_cell::sync::Lazy<Mutex<GlobalProfilerUi>> =
     once_cell::sync::Lazy::new(Default::default);
 
 /// Show the profiler.
@@ -120,28 +122,155 @@ pub fn profiler_ui(ui: &mut egui::Ui) {
     profile_ui.ui(ui);
 }
 
-fn latest_frames() -> Frames {
-    let profiler = GlobalProfiler::lock();
-    Frames {
-        recent: profiler.recent_frames().cloned().collect(),
-        slowest: profiler.slowest_frames_chronological(),
+// ----------------------------------------------------------------------------
+
+/// Show [`puffin::GlobalProfiler`], i.e. profile the app we are running in.
+#[derive(Default)]
+#[cfg_attr(feature = "serde", derive(serde::Deserialize, serde::Serialize))]
+#[cfg_attr(feature = "serde", serde(default))]
+pub struct GlobalProfilerUi {
+    #[cfg_attr(feature = "serde", serde(skip))]
+    global_frame_view: GlobalFrameView,
+    pub profiler_ui: ProfilerUi,
+}
+
+impl GlobalProfilerUi {
+    /// Show an [`egui::Window`] with the profiler contents.
+    ///
+    /// If you want to control the window yourself, use [`Self::ui`] instead.
+    ///
+    /// Returns `false` if the user closed the profile window.
+    pub fn window(&mut self, ctx: &egui::CtxRef) -> bool {
+        let mut frame_view = self.global_frame_view.lock();
+        self.profiler_ui.window(ctx, &mut frame_view)
+    }
+
+    /// Show the profiler.
+    ///
+    /// Call this from within an [`egui::Window`], or use [`Self::window`] instead.
+    pub fn ui(&mut self, ui: &mut egui::Ui) {
+        let mut frame_view = self.global_frame_view.lock();
+        self.profiler_ui.ui(ui, &mut frame_view);
     }
 }
 
 // ----------------------------------------------------------------------------
 
-/// The frames we can select between
+/// The frames we can chose between when selecting what frame(s) to view.
 #[derive(Clone)]
-pub struct Frames {
+pub struct AvailableFrames {
     pub recent: Vec<Arc<FrameData>>,
     pub slowest: Vec<Arc<FrameData>>,
 }
 
+impl AvailableFrames {
+    fn latest(frame_view: &FrameView) -> Self {
+        Self {
+            recent: frame_view.recent_frames().cloned().collect(),
+            slowest: frame_view.slowest_frames_chronological(),
+        }
+    }
+}
+
+/// Multiple streams for one thread.
+#[derive(Clone)]
+pub struct Streams {
+    streams: Vec<Arc<StreamInfo>>,
+    merged_scopes: Vec<MergeScope<'static>>,
+    max_depth: usize,
+}
+
+impl Streams {
+    fn new(frames: &[Arc<FrameData>], thread_info: &ThreadInfo) -> Self {
+        crate::profile_function!();
+
+        let mut streams = vec![];
+        for frame in frames {
+            if let Some(stream_info) = frame.thread_streams.get(thread_info) {
+                streams.push(stream_info.clone());
+            }
+        }
+
+        let merges = puffin::merge_scopes_for_thread(frames, thread_info).unwrap();
+        let merges = merges.into_iter().map(|ms| ms.into_owned()).collect();
+
+        let mut max_depth = 0;
+        for stream_info in &streams {
+            max_depth = stream_info.depth.max(max_depth);
+        }
+
+        Self {
+            streams,
+            merged_scopes: merges,
+            max_depth,
+        }
+    }
+}
+
+/// Selected frames ready to be viewed.
+/// Never empty.
+#[derive(Clone)]
+pub struct SelectedFrames {
+    /// ordered, but not necessarily in sequence
+    pub frames: vec1::Vec1<Arc<FrameData>>,
+    pub raw_range_ns: (NanoSecond, NanoSecond),
+    pub merged_range_ns: (NanoSecond, NanoSecond),
+    pub threads: BTreeMap<ThreadInfo, Streams>,
+}
+
+impl SelectedFrames {
+    fn try_from_vec(frames: Vec<Arc<FrameData>>) -> Option<Self> {
+        let frames = vec1::Vec1::try_from_vec(frames).ok()?;
+        Some(Self::from_vec1(frames))
+    }
+
+    fn from_vec1(mut frames: vec1::Vec1<Arc<FrameData>>) -> Self {
+        frames.sort_by_key(|f| f.frame_index);
+        frames.dedup_by_key(|f| f.frame_index);
+
+        let mut threads: BTreeSet<ThreadInfo> = BTreeSet::new();
+        for frame in &frames {
+            for ti in frame.thread_streams.keys() {
+                threads.insert(ti.clone());
+            }
+        }
+
+        let threads: BTreeMap<ThreadInfo, Streams> = threads
+            .iter()
+            .map(|ti| (ti.clone(), Streams::new(&frames, ti)))
+            .collect();
+
+        let mut merged_min_ns = NanoSecond::MAX;
+        let mut merged_max_ns = NanoSecond::MIN;
+        for stream in threads.values() {
+            for scope in &stream.merged_scopes {
+                let scope_start = scope.relative_start_ns;
+                let scope_end = scope_start + scope.duration_per_frame_ns;
+                merged_min_ns = merged_min_ns.min(scope_start);
+                merged_max_ns = merged_max_ns.max(scope_end);
+            }
+        }
+
+        let raw_range_ns = (frames.first().range_ns.0, frames.last().range_ns.1);
+
+        Self {
+            frames,
+            raw_range_ns,
+            merged_range_ns: (merged_min_ns, merged_max_ns),
+            threads,
+        }
+    }
+
+    pub fn contains(&self, frame_index: u64) -> bool {
+        self.frames.iter().any(|f| f.frame_index == frame_index)
+    }
+}
+
 #[derive(Clone)]
 pub struct Paused {
     /// What we are viewing
-    selected: Arc<FrameData>,
-    frames: Frames,
+    selected: SelectedFrames,
+    frames: AvailableFrames,
 }
 
 #[derive(Copy, Clone, Debug, PartialEq)]
@@ -181,50 +310,50 @@ impl ProfilerUi {
     /// If you want to control the window yourself, use [`Self::ui`] instead.
     ///
     /// Returns `false` if the user closed the profile window.
-    pub fn window(&mut self, ctx: &egui::CtxRef) -> bool {
+    pub fn window(&mut self, ctx: &egui::CtxRef, frame_view: &mut FrameView) -> bool {
         puffin::profile_function!();
         let mut open = true;
         egui::Window::new("Profiler")
             .default_size([1024.0, 600.0])
             .open(&mut open)
-            .show(ctx, |ui| self.ui(ui));
+            .show(ctx, |ui| self.ui(ui, frame_view));
         open
     }
 
     /// The frames we can select between
-    fn frames(&self) -> Frames {
-        self.paused
-            .as_ref()
-            .map_or_else(latest_frames, |paused| paused.frames.clone())
+    fn frames(&self, frame_view: &FrameView) -> AvailableFrames {
+        self.paused.as_ref().map_or_else(
+            || AvailableFrames::latest(frame_view),
+            |paused| paused.frames.clone(),
+        )
     }
 
     /// Pause on the specific frame
-    fn pause_and_select(&mut self, selected: Arc<FrameData>) {
+    fn pause_and_select(&mut self, frame_view: &FrameView, selected: SelectedFrames) {
         if let Some(paused) = &mut self.paused {
             paused.selected = selected;
         } else {
             self.paused = Some(Paused {
                 selected,
-                frames: self.frames(),
+                frames: self.frames(frame_view),
             });
         }
     }
 
-    fn selected_frame(&self) -> Option<Arc<FrameData>> {
-        self.paused
-            .as_ref()
-            .map(|paused| paused.selected.clone())
-            .or_else(|| GlobalProfiler::lock().latest_frame())
-    }
-
-    fn selected_frame_index(&self) -> Option<FrameIndex> {
-        self.selected_frame().map(|frame| frame.frame_index)
+    fn is_selected(&self, frame_view: &FrameView, frame_index: u64) -> bool {
+        if let Some(paused) = &self.paused {
+            paused.selected.contains(frame_index)
+        } else if let Some(latest_frame) = frame_view.latest_frame() {
+            latest_frame.frame_index == frame_index
+        } else {
+            false
+        }
     }
 
     /// Show the profiler.
     ///
     /// Call this from within an [`egui::Window`], or use [`Self::window`] instead.
-    pub fn ui(&mut self, ui: &mut egui::Ui) {
+    pub fn ui(&mut self, ui: &mut egui::Ui, frame_view: &mut FrameView) {
         #![allow(clippy::collapsible_else_if)]
         puffin::profile_function!();
 
@@ -238,15 +367,24 @@ impl ProfilerUi {
         egui::CollapsingHeader::new("Frames")
             .default_open(true)
             .show(ui, |ui| {
-                hovered_frame = self.show_frames(ui);
+                hovered_frame = self.show_frames(ui, frame_view);
             });
 
-        let frame = match hovered_frame.or_else(|| self.selected_frame()) {
-            Some(frame) => frame,
-            None => {
-                ui.label("No profiling data");
-                return;
-            }
+        let frames = if let Some(hovered_frame) = hovered_frame {
+            SelectedFrames::try_from_vec(vec![hovered_frame])
+        } else if let Some(paused) = &self.paused {
+            Some(paused.selected.clone())
+        } else if let Some(latest_frame) = frame_view.latest_frame() {
+            SelectedFrames::try_from_vec(vec![latest_frame])
+        } else {
+            None
+        };
+
+        let frames = if let Some(frames) = frames {
+            frames
+        } else {
+            ui.label("No profiling data");
+            return;
         };
 
         // TODO: show age of data
@@ -270,23 +408,18 @@ impl ProfilerUi {
                         .clicked()
                         || ui.input().key_pressed(egui::Key::Space)
                     {
-                        let latest = GlobalProfiler::lock().latest_frame();
+                        let latest = frame_view.latest_frame();
                         if let Some(latest) = latest {
-                            self.pause_and_select(latest);
+                            self.pause_and_select(
+                                frame_view,
+                                SelectedFrames::from_vec1(vec1::vec1![latest]),
+                            );
                         }
                     }
                 });
             }
             ui.separator();
-            let (min_ns, max_ns) = frame.range_ns;
-            ui.label(format!(
-                "Showing frame #{}, {:.1} ms, {} threads, {} scopes, {:.1} kB",
-                frame.frame_index,
-                (max_ns - min_ns) as f64 * 1e-6,
-                frame.thread_streams.len(),
-                frame.num_scopes,
-                frame.num_bytes as f64 * 1e-3
-            ));
+            frames_info_ui(ui, &frames);
         });
 
         if self.paused.is_none() {
@@ -303,24 +436,38 @@ impl ProfilerUi {
         ui.separator();
 
         match self.view {
-            View::Flamegraph => flamegraph::ui(ui, &mut self.options, &frame),
-            View::Stats => stats::ui(ui, &frame),
+            View::Flamegraph => flamegraph::ui(ui, &mut self.options, &frames),
+            View::Stats => stats::ui(ui, &frames.frames),
         }
     }
 
     /// Returns hovered, if any
-    fn show_frames(&mut self, ui: &mut egui::Ui) -> Option<Arc<FrameData>> {
-        let frames = self.frames();
+    fn show_frames(
+        &mut self,
+        ui: &mut egui::Ui,
+        frame_view: &mut FrameView,
+    ) -> Option<Arc<FrameData>> {
+        let frames = self.frames(frame_view);
 
         let mut hovered_frame = None;
 
         let longest_count = frames.recent.len().max(frames.slowest.len());
 
         egui::Grid::new("frame_grid").num_columns(2).show(ui, |ui| {
+            ui.label("");
+            ui.label("Click to select a frame, or drag to select multiple frames.");
+            ui.end_row();
+
             ui.label("Recent:");
 
             Frame::dark_canvas(ui.style()).show(ui, |ui| {
-                self.show_frame_list(ui, &frames.recent, longest_count, &mut hovered_frame);
+                self.show_frame_list(
+                    ui,
+                    frame_view,
+                    &frames.recent,
+                    longest_count,
+                    &mut hovered_frame,
+                );
             });
 
             ui.end_row();
@@ -330,11 +477,17 @@ impl ProfilerUi {
                 ui.add_space(16.0); // make it a bit more centered
                 ui.label("Slowest:");
                 if ui.button("Clear").clicked() {
-                    GlobalProfiler::lock().clear_slowest();
+                    frame_view.clear_slowest();
                 }
             });
             Frame::dark_canvas(ui.style()).show(ui, |ui| {
-                self.show_frame_list(ui, &frames.slowest, longest_count, &mut hovered_frame);
+                self.show_frame_list(
+                    ui,
+                    frame_view,
+                    &frames.slowest,
+                    longest_count,
+                    &mut hovered_frame,
+                );
             });
         });
 
@@ -344,6 +497,7 @@ impl ProfilerUi {
     fn show_frame_list(
         &mut self,
         ui: &mut egui::Ui,
+        frame_view: &FrameView,
         frames: &[Arc<FrameData>],
         longest_count: usize,
         hovered_frame: &mut Option<Arc<FrameData>>,
@@ -357,7 +511,7 @@ impl ProfilerUi {
             ui.available_size_before_wrap_finite().x,
             self.options.frame_list_height,
         );
-        let (response, painter) = ui.allocate_painter(desired_size, Sense::drag());
+        let (response, painter) = ui.allocate_painter(desired_size, Sense::click_and_drag());
         let rect = response.rect;
 
         let frame_width_including_spacing =
@@ -365,7 +519,13 @@ impl ProfilerUi {
         let frame_spacing = 2.0;
         let frame_width = frame_width_including_spacing - frame_spacing;
 
-        let selected_frame_index = self.selected_frame_index();
+        let viewing_multiple_frames = if let Some(paused) = &self.paused {
+            paused.selected.frames.len() > 1 && !self.options.merge_scopes
+        } else {
+            false
+        };
+
+        let mut new_selection = vec![];
 
         for (i, frame) in frames.iter().enumerate() {
             let x = rect.right() - (frames.len() as f32 - i as f32) * frame_width_including_spacing;
@@ -376,10 +536,11 @@ impl ProfilerUi {
 
             let duration = frame.duration_ns();
 
-            let is_selected = Some(frame.frame_index) == selected_frame_index;
+            let is_selected = self.is_selected(frame_view, frame.frame_index);
 
             let is_hovered = if let Some(mouse_pos) = response.hover_pos() {
                 response.hovered()
+                    && !response.dragged()
                     && frame_rect
                         .expand2(vec2(0.5 * frame_spacing, 0.0))
                         .contains(mouse_pos)
@@ -387,14 +548,26 @@ impl ProfilerUi {
                 false
             };
 
-            if is_hovered {
+            // preview when hovering is really annoying when viewing multiple frames
+            if is_hovered && !is_selected && !viewing_multiple_frames {
                 *hovered_frame = Some(frame.clone());
                 egui::show_tooltip_at_pointer(ui.ctx(), Id::new("puffin_frame_tooltip"), |ui| {
                     ui.label(format!("{:.1} ms", frame.duration_ns() as f64 * 1e-6));
                 });
             }
-            if is_hovered && response.clicked() {
-                self.pause_and_select(frame.clone());
+
+            if response.dragged() {
+                if let (Some(start), Some(curr)) = (
+                    ui.input().pointer.press_origin(),
+                    ui.input().pointer.interact_pos(),
+                ) {
+                    let min_x = start.x.min(curr.x);
+                    let max_x = start.x.max(curr.x);
+                    let intersects = min_x <= frame_rect.right() && frame_rect.left() <= max_x;
+                    if intersects {
+                        new_selection.push(frame.clone());
+                    }
+                }
             }
 
             let color = if is_selected {
@@ -417,5 +590,47 @@ impl ProfilerUi {
             );
             painter.rect_filled(short_rect, 0.0, color);
         }
+
+        if let Some(new_selection) = SelectedFrames::try_from_vec(new_selection) {
+            self.pause_and_select(frame_view, new_selection);
+        }
     }
 }
+
+fn frames_info_ui(ui: &mut egui::Ui, selection: &SelectedFrames) {
+    let mut sum_ns = 0;
+    let mut sum_scopes = 0;
+    let mut sum_bytes = 0;
+
+    for frame in &selection.frames {
+        let (min_ns, max_ns) = frame.range_ns;
+        sum_ns += max_ns - min_ns;
+
+        sum_scopes += frame.num_scopes;
+        sum_bytes += frame.num_bytes;
+    }
+
+    let frame_indices = if selection.frames.len() == 1 {
+        format!("frame #{}", selection.frames[0].frame_index)
+    } else if selection.frames.len() as u64
+        == selection.frames.last().frame_index - selection.frames.first().frame_index + 1
+    {
+        format!(
+            "{} frames (#{} - #{})",
+            selection.frames.len(),
+            selection.frames.first().frame_index,
+            selection.frames.last().frame_index
+        )
+    } else {
+        format!("{} frames", selection.frames.len())
+    };
+
+    ui.label(format!(
+        "Showing {}, {:.1} ms, {} threads, {} scopes, {:.1} kB",
+        frame_indices,
+        sum_ns as f64 * 1e-6,
+        selection.threads.len(),
+        sum_scopes,
+        sum_bytes as f64 * 1e-3
+    ));
+}
